# -*- coding: utf-8 -*-


import json
import functools
import itertools
import collections


class TagManager(object):

    def __init__(self, registrations):
        self._registrations = registrations
        self._mappings = []

    def infere_tags(self, reference):
        result = set()
        result.add(reference)

        if not self._registrations.get(reference) or not self._registrations[reference].get('dependencies'):
            return result

        for dependency_reference in self._registrations[reference]['dependencies']:

            dependency = self._registrations.get(dependency_reference)
            # if not dependency: continue

            dependency_tags = dependency.get('tags')
            # if not dependency_tags: continue

            for tag in dependency_tags:
                inherited_tag = ('{%s}' % tag) if tag[0] != '{' or tag[-1] != '}' else tag
                result.add(inherited_tag)

        return result

    def register(self, fn, *selections):
        if len(selections) is 0:
            return

        self._mappings.append((fn, selections))
        self._evaluate_current_selections(fn, selections)

    def register_synched(self, fn, selections, references=[]):
        if len(selections) is 0:
            self._execute_fn(fn, *references)
            return

        selection = selections[0]
        for reference in references:
            selection = selection.intersection('{%s}' % reference)

        def _wrapper(reference):
            self.register_synched(fn, selections[1:], references + [reference])

        self.register(_wrapper, selection)

    def evaluate_new_candidate(self, new_reference):
        for fn, selections in self._mappings:
            self._evaluate_existing_selections(fn, selections, new_reference)

    def _evaluate_existing_selections(self, fn, selections, new_reference):
        partial_elegible_references_list = []
        current_partial_references_list = []

        for selection in selections:
            partial_references_list_copy = None

            if selection.is_elegible(new_reference):
                partial_references_list_copy = current_partial_references_list[:]
                partial_references_list_copy.append([new_reference])

            references = [reference for reference in selection]
            current_partial_references_list.append(references)
            for partial_elegible_references in partial_elegible_references_list:
                partial_elegible_references.append(references)

            if partial_references_list_copy:
                partial_elegible_references_list.append(partial_references_list_copy)

        for elegible_references_list in partial_elegible_references_list:
            combinations = itertools.product(*elegible_references_list)
            for combination in combinations:
                self._execute_fn(fn, *combination)

    def _evaluate_current_selections(self, fn, selections):
        references = [[reference for reference in selection] for selection in selections]
        if len(references) is 0:
            return

        combinations = itertools.product(*references)
        for combination in combinations:
            self._execute_fn(fn, *combination)

    def _execute_fn(self, fn, *references):
        fn(*references)


class TagSelector(object):

    def __init__(self, callable_factory):
        self._callable_factory = callable_factory
        self._tag_mappings = collections.defaultdict(set)
        self._reference_mappings = collections.defaultdict(set)

    def register(self, reference, tags):
        self._reference_mappings[reference] = tags
        for tag in tags:
            self._tag_mappings[tag].add(reference)

    def get(self, *selectors):
        selection = _TagSelection(self._tag_mappings, self._callable_factory)
        return selection if len(selectors) is 0 else selection.union(*selectors)

    def tags(self, *references):
        selection = _TagSelection(self._reference_mappings, self.get)
        return selection if len(references) is 0 else selection.union(*references)


class _TagSelection(object):

    def __init__(self, tag_mappings, callable_factory, operations=[], initial=[]):
        self._tag_mappings = tag_mappings
        self._callable_factory = callable_factory
        self._operations = operations
        self._initial_selection = initial

    def __iter__(self):
        return iter(self._selection)

    def __str__(self):
        return self._to_string()

    def __repr__(self):
        return self._to_string()

    def __len__(self):
        return len(self._selection)

    def is_elegible(self, reference):
        return reference in self._selection

    def get(self, index=None):
        if index is not None:
            return self._callable_factory(sorted(self._selection)[index])

        result = {}
        for selected in sorted(self._selection):
            result[selected] = self._callable_factory(selected)
        return result

    def all(self):
        return _TagSelection(self._tag_mappings,
                             self._callable_factory,
                             [self._evaluate_all],
                             self._initial_selection)

    def union(self, *selectors):
        operation = functools.partial(self._evaluate_union, selectors)
        return _TagSelection(self._tag_mappings,
                             self._callable_factory,
                             self._operations + [operation],
                             self._initial_selection)

    def intersection(self, *selectors):
        operation = functools.partial(self._evaluate_intersection, selectors)
        return _TagSelection(self._tag_mappings,
                             self._callable_factory,
                             self._operations + [operation],
                             self._initial_selection)

    def difference(self, *selectors):
        operation = functools.partial(self._evaluate_difference, selectors)
        return _TagSelection(self._tag_mappings,
                             self._callable_factory,
                             self._operations + [operation],
                             self._initial_selection)

    def symmetric_difference(self, *selectors):
        operation = functools.partial(self._evaluate_symmetric_difference, selectors)
        return _TagSelection(self._tag_mappings,
                             self._callable_factory,
                             self._operations + [operation],
                             self._initial_selection)

    @property
    def _selection(self):
        selection = set(self._initial_selection)
        for operation in self._operations:
            selection = operation(selection)
        return selection

    def _evaluate_all(self, selection):
        group = set()
        for values in self._tag_mappings.itervalues():
            group |= values
        return group

    def _evaluate_union(self, selectors, selection):
        selection |= self._evaluate_selectors(*selectors)
        return selection

    def _evaluate_intersection(self, selectors, selection):
        selection &= self._evaluate_selectors(*selectors)
        return selection

    def _evaluate_difference(self, selectors, selection):
        selection -= self._evaluate_selectors(*selectors)
        return selection

    def _evaluate_symmetric_difference(self, selectors, selection):
        selection ^= self._evaluate_selectors(*selectors)
        return selection

    def _evaluate_selectors(self, *selectors):
        if len(selectors) is 0:
            return set()

        group = self._tag_mappings[selectors[0]].copy()
        for selector in selectors[1:]:
            group &= self._tag_mappings[selector]
        return group

    def _to_string(self):
        return '<selection %s currently-with="%s">' % (id(self), json.dumps(list(self._selection), indent=4))



# -*- coding: utf-8 -*-


import bigtempo.utils as utils
import bigtempo.defaults as defaults
import bigtempo.tagselection as tagselection


class DatasourceEngine(object):

    def __init__(self,
                 datasource_factory=defaults.datasource_factory,
                 processing_task_factory=defaults.task_factory):

        self._registrations = {}
        self._instances = {}
        self._tag_selector = tagselection.TagSelector(self.get)
        self._tag_iteration_manager = tagselection.TagManager(self._registrations)
        self._processing_task_factory = processing_task_factory
        self._datasource_factory = datasource_factory

    def for_each(self, *selection):
        def wrapper(fn):
            self._tag_iteration_manager.register(fn, *selection)
            return fn
        return wrapper

    def for_synched(self, *selection):
        def wrapper(fn):
            self._tag_iteration_manager.register_synched(fn, selection)
            return fn
        return wrapper

    def datasource(self, reference, dependencies=None, tags=None, lookback=0, frequency=None):
        def wrapper(cls):
            self._register_datasource(reference,
                                      cls,
                                      dependencies=dependencies,
                                      tags=tags,
                                      lookback=lookback,
                                      frequency=frequency)
            return cls
        return wrapper

    def _register_datasource(self, reference, cls, dependencies=None, tags=None, lookback=0, frequency=None):
        self._instances[reference] = None
        self._registrations[reference] = {
            'class': cls,
            'lookback': lookback,
            'frequency': frequency,
            'dependencies': set(dependencies) if dependencies else set()
        }

        infered_tags = self._tag_iteration_manager.infere_tags(reference)
        all_tags = utils.assure_is_valid_set(tags) | utils.assure_is_valid_set(infered_tags)

        self._registrations[reference]['tags'] = all_tags
        self._tag_selector.register(reference, all_tags)
        self._tag_iteration_manager.evaluate_new_candidate(reference)

    def select(self, *selectors):
        return self._tag_selector.get(*selectors)

    def tags(self, *references):
        return self._tag_selector.tags(*references)

    def get(self, reference):
        return self._create_processor(reference)

    def _create_processor(self, reference):
        instance = self._lazyload_datasource(reference)
        registration = self._registrations[reference]
        dependencies = dict((dependency, self._create_processor(dependency)) for dependency in registration['dependencies'])
        return self._processing_task_factory(instance, registration, dependencies)

    def _lazyload_datasource(self, reference):
        if not self._instances.get(reference):
            if not self._registrations.get(reference):
                raise KeyError()
            self._instances[reference] = self._datasource_factory(self._registrations[reference]['class'])
        return self._instances[reference]



# -*- coding: utf-8 -*-


import pandas
import datetime

import bigtempo.utils as utils


_DEFAULT_FREQUENCY = 'B'
_KNOWN_FREQUENCIES = ['U', 'L', 'S', 'T', 'H', 'B', 'W', 'BMS', 'BM', 'MS', 'M', 'BQS', 'BQ', 'QS', 'Q', 'BAS', 'BA', 'AS', 'A']
_FREQUENCY_ENUMERATION_DICT = dict((y, x) for x, y in enumerate(_KNOWN_FREQUENCIES))


def factory(instance, registration, dependency_dict, *args, **kwargs):
    return DataFrameDatasourceTask(instance, registration, dependency_dict)


class DataFrameDatasourceTask(object):

    def __init__(self, instance, registration, dependency_dict):
        self._instance = instance
        self._dependency_dict = dependency_dict
        self._registration = registration

    def process(self, symbol, start=None, end=None):
        context = self._create_context_for(symbol, start, end)
        result = self._instance.evaluate(context, symbol, start, end)
        return utils.slice(result, start, end)

    def _create_context_for(self, symbol, start=None, end=None):
        evaluated_dependencies = self._evaluate_datasource_dependencies(symbol, start, end)
        return DatasourceContext(evaluated_dependencies)

    def _evaluate_datasource_dependencies(self, symbol, start=None, end=None):
        result = {}

        new_start = None if not start else evaluate_loopback_period(self._registration,
                                                                    self._dependency_dict.values(),
                                                                    start)

        for reference, dependency in self._dependency_dict.iteritems():
            result[reference] = dependency.process(symbol, new_start, end)
        return result


class DatasourceContext(object):

    def __init__(self, dependencies):
        self._dependencies = dependencies

    def dependencies(self, reference=None):
        return self._dependencies.get(reference) if reference else self._dependencies


def evaluate_loopback_period(datasource_registration, dependencies, date):
    lookback = datasource_registration['lookback']
    frequency = determine_frequency(datasource_registration.get('frequency'), dependencies)

    # Holiday workaround
    if frequency in ['B', 'C']:
        lookback = 1 + int(lookback * 1.08)

    lookback += 1

    return relative_period(-lookback, frequency, date)


def determine_frequency(datasource_frequency=None, dependencies=None):
    if not datasource_frequency is None:
        return datasource_frequency

    if dependencies is None or len(dependencies) is 0:
        return _DEFAULT_FREQUENCY

    dependencies_frequencies = []
    for dependency in dependencies:
        dependency_frequency = dependency._registration.get('frequency')
        dependency_dependencies = dependency._dependency_dict.values()
        dependencies_frequencies.append(determine_frequency(dependency_frequency, dependency_dependencies))

    return max(dependencies_frequencies, key=_frequency_sort_key)


def _frequency_sort_key(value):
    frequency = value.split('-')[0]
    if not frequency in _FREQUENCY_ENUMERATION_DICT:
        return 0

    return _FREQUENCY_ENUMERATION_DICT[frequency]


def relative_period(periods, frequency, date=None):
    business_day = equivalent_business_day() if not date else equivalent_business_day(date)
    return (pandas.Period(business_day, freq=frequency) + periods).to_timestamp()


def equivalent_business_day(date=None):
    if not date:
        date = datetime.datetime.today().replace(hour=0,
                                                 minute=0,
                                                 second=0,
                                                 microsecond=0)

    isoweekday = date.isoweekday()
    return date if isoweekday <= 5 else date - datetime.timedelta(isoweekday - 5)



# -*- coding: utf-8 -*-


class CallableMock(object):
    def __init__(self, mock):
        self.mock = mock

    def __call__(self, *args, **kwargs):
        return self.mock.__call__(*args, **kwargs)


class IterableMock(object):
    def __init__(self, mock):
        self.mock = mock

    def __iter__(self, *args, **kwargs):
        return self.mock.__iter__(*args, **kwargs)

    def __getattr__(self, method_name):
        return self.mock.__getattr__(method_name)


class DatasourceLogger(object):

    def __init__(self):
        self.logs = []

    def log(self, description, dataframe):
        self.logs.append((description, dataframe))

    def print_summary(self):
        for description, dataframe in self.logs:
            title = description + ' ' + ((80 - len(description) - 1) * '-')
            size = len(dataframe.index)
            start = '-------------------' if size is 0 else dataframe.index[0]
            end = '-------------------' if size is 0 else dataframe.index[-1]
            print '[ %s : %s ] ~ %i \t > \t %s' % (start, end, size, title)

        for description, dataframe in self.logs:
            print ''
            title = description + ' ' + ((80 - len(description) - 1) * '-')
            print title
            print dataframe.head(2)
            print dataframe.tail(2)


def assure_is_valid_set(obj):
    if not obj:
        obj = set()
    elif not isinstance(obj, set):
        try:
            obj = set(obj)
        except:
            obj = set()
    return obj


def slice(slicable, start=None, end=None):
    if start and end:
        return slicable[start:end]
    if start:
        return slicable[start:]
    if end:
        return slicable[:end]
    return slicable



# -*- coding: utf-8 -*-


import bigtempo.processors.dataframe_task as task


def sample_datasource_factory(cls):
    return cls()


datasource_factory = sample_datasource_factory
task_factory = task.factory



#!/usr/bin/env python
# Filename: setup.py


import os
import sys


if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    sys.exit()


import re
import bigtempo
from pkgutil import walk_packages


try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup


def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()


def filter_comments(contents):
    filter_pattern = re.compile(r'[\s]*#.*')
    return filter(lambda x: not filter_pattern.match(x), contents)


def packages(path=None, prefix="", exclude=None):
    try:
        return find_packages(exclude=exclude)
    except:
        return [name for _, name, ispkg in walk_packages(path, prefix) if ispkg]


setup(
    name='bigtempo',
    version=bigtempo.__version__,
    description='Powerful processment of temporal data.',
    long_description=read('README.rst'),
    license=read('LICENSE'),

    author='Roberto Haddock Lobo',
    author_email='rhlobo+bigtempo@gmail.com',
    url='https://github.com/rhlobo/bigtempo',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Financial and Insurance Industry',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],

    install_requires=filter_comments(read('requirements.txt').split('\n')),
    package_dir={'bigtempo': 'bigtempo'},
    packages=packages(bigtempo.__path__,
                      bigtempo.__name__,
                      exclude=["*.tests",
                               "*.tests.*",
                               "tests.*",
                               "tests"]),
    package_data={'': ['AUTHORS.rst',
                       'README.rst',
                       'LICENSE',
                       'requirements.txt']},

    include_package_data=True,
    zip_safe=False
)



# -*- coding: utf-8 -*-


def factory(instance, registration, dependencies, *args, **kwargs):
    return SimpleDatasourceTask(instance, registration, dependencies)


class SimpleDatasourceTask(object):

    def __init__(self, instance, registration, dependencies):
        self._instance = instance
        self._dependencies = dependencies

    def process(self, **kwargs):
        context = self._evaluate_datasource_dependencies(**kwargs)
        return self._instance.evaluate(context, **kwargs)

    def _evaluate_datasource_dependencies(self, **kwargs):
        return dict((reference, dependency.process(**kwargs)) for reference, dependency in self._dependencies.iteritems())



# -*- coding: utf-8 -*-


import os
import copy
import pandas

import unittest
from mockito import mock, unstub

import bigtempo.utils as utils


def generate_multiple(engine, references, symbol, start, end, test_data_filepath_fn, module=None):
    results = []
    for reference in references:
        result = generate(engine, reference, symbol, start, end, test_data_filepath_fn, module)
        results.append(result)
    if module is None:
        return results


def generate(engine, reference, symbol, start, end, test_data_filepath_fn, module=None):
    if not os.path.isfile(test_data_filepath_fn(reference, symbol)):
        return

    class CustomDatasourceTestCase(_create_datasource_test_case_for(engine)):

        def test_datasource_using_test_data(self):
            _assert_datasource_correctness_using_datafiles(engine, reference, symbol, start, end, test_data_filepath_fn)

    CustomDatasourceTestCase.__name__ = '%s{%s}[%s:%s]' % (reference, symbol, start, end)

    if module is None:
        return CustomDatasourceTestCase

    setattr(module,
            'TestDatasourceCorrectness(%s){%s}[%s:%s]' % (reference, symbol, start, end),
            CustomDatasourceTestCase)


def _create_datasource_test_case_for(engine):
    class DatasourceTestCase(unittest.TestCase):

        @classmethod
        def setUpClass(cls):
            testing_datasource_factory_mock = mock()

            def testing_datasource_factory(cls):
                mock_result = testing_datasource_factory_mock.build(cls)
                if mock_result:
                    return mock_result
                return cls()

            cls.testing_datasource_factory_mock = testing_datasource_factory_mock
            cls.original_datasource_factory = engine._datasource_factory
            engine._datasource_factory = testing_datasource_factory

        @classmethod
        def tearDownClass(cls):
            engine._datasource_factory = cls.original_datasource_factory
            engine._instances = {}

        def setUp(self):
            engine._instances = {}

        def tearDown(self):
            unstub()

    return DatasourceTestCase


def _assert_datasource_correctness_using_datafiles(engine, reference, symbol, start, end, test_data_filepath_fn):
    logger = utils.DatasourceLogger()

    original_registrations = {}

    for dependency_reference in engine._registrations[reference]['dependencies']:
        mock_data_file = test_data_filepath_fn(dependency_reference, symbol)

        if not os.path.isfile(mock_data_file):
            continue

        datasource_mock_cls = _create_mock_datasource(dependency_reference, mock_data_file, logger)

        if not engine._registrations.get(dependency_reference):
            engine._registrations[dependency_reference] = {}

        original_registrations[dependency_reference] = engine._registrations[dependency_reference]
        registration_mock = copy.deepcopy(engine._registrations[dependency_reference])
        registration_mock['class'] = datasource_mock_cls
        registration_mock['dependencies'] = set()
        engine._registrations[dependency_reference] = registration_mock

    expected_data_file = test_data_filepath_fn(reference, symbol)

    actual = engine.get(reference).process(symbol, start, end)
    expected = utils.slice(pandas.DataFrame.from_csv(expected_data_file), start, end)

    logger.log('actual', actual)
    logger.log('expected', expected)

    try:
        _assert_dataframe_almost_equal(expected, actual)
    except Exception, e:
        logger.print_summary()
        raise e
    finally:
        for dependency_reference, original_registration in original_registrations.items():
            engine._registrations[dependency_reference] = original_registration


def _create_mock_datasource(mock_reference, mock_data_file, logger):
    class DatasourceMock(object):

        def evaluate(self, context, symbol, start=None, end=None):
            data = pandas.DataFrame.from_csv(mock_data_file)
            sliced_data = utils.slice(data, start, end)
            logger.log(os.path.basename(mock_data_file), sliced_data)
            return sliced_data

    return DatasourceMock


def _assert_dataframe_almost_equal(expected, actual, margin=0.0000000001):
    tmp = ((expected.dropna() - actual.dropna()).abs() < margin)
    assert tmp.all().all()
